"""
To run the tests:

python -m unittest tests.test_live_staging.TestLiveStaging.test_workflows_list -v
"""

import json
import os
import time
import unittest
import random
import string
from pathlib import Path
from dotenv import load_dotenv

from cloudcruise import CloudCruise, CloudCruiseParams
from cloudcruise.runs.types import StartRunRequest

# Load .env located in the same directory as this test file
load_dotenv(Path(__file__).parent / ".env")

API_KEY = os.environ.get("CLOUDCRUISE_API_KEY") or None
ENCRYPTION_KEY = os.environ.get("CLOUDCRUISE_ENCRYPTION_KEY") or None
BASE_URL = os.environ.get("CLOUDCRUISE_BASE_URL") or "https://api.cloudcruise.com"

# Vault test parameters
VAULT_DOMAIN = "https://example.com"
VAULT_USER_ID = "user123"

# Runs test parameters
RUN_WORKFLOW_ID = None  # e.g., "workflow-123"
RUN_INPUT_JSON = None  # e.g., '{"url": "https://example.com"}'


class TestLiveStaging(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        if not API_KEY or not ENCRYPTION_KEY:
            raise unittest.SkipTest("Provide CLOUDCRUISE_API_KEY and CLOUDCRUISE_ENCRYPTION_KEY or set in file")

        cls.client = CloudCruise(CloudCruiseParams(api_key=API_KEY, base_url=BASE_URL, encryption_key=ENCRYPTION_KEY))

    @unittest.skipIf(True, "workflows test skipped")
    def test_workflows_list(self):
        workflows = self.client.workflows.get_all_workflows()
        self.assertIsInstance(workflows, list)

    @unittest.skipIf(True, "vault test skipped")
    def test_vault_crud(self):
        if not VAULT_DOMAIN or not VAULT_USER_ID:
            self.skipTest("Set VAULT_DOMAIN and VAULT_USER_ID in this file to run")

        alias = "sdk-live-" + "".join(random.choices(string.ascii_lowercase + string.digits, k=8))
        entry = self.client.vault.create(VAULT_DOMAIN, VAULT_USER_ID, {
            "user_name": "test_user",
            "password": "test_password",
            "user_alias": alias,
        })
        self.assertTrue(entry.get("id"))

        # Get (filtered, decrypted)
        entries = self.client.vault.get()
        self.assertIsInstance(entries, list)

        from cloudcruise.vault.types import GetVaultEntriesFilters
        filt = self.client.vault.get(GetVaultEntriesFilters(
            permissioned_user_id=VAULT_USER_ID, domain=VAULT_DOMAIN
        ))
        self.assertIsInstance(filt, list)

        # Update
        updated = self.client.vault.update({
            "permissioned_user_id": VAULT_USER_ID,
            "user_name": "test_user",
            "password": "test_password",
            "domain": VAULT_DOMAIN,
            "user_alias": alias + "-u",
        })
        self.assertEqual(updated.get("user_alias"), alias + "-u")

        # Delete
        self.client.vault.delete({"domain": VAULT_DOMAIN, "permissioned_user_id": VAULT_USER_ID})

if __name__ == "__main__":
    unittest.main(verbosity=2)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional

from ..runs.types import EventType
from ..events.types import WebhookMessage


class VerificationError(Exception):
    def __init__(self, message: str = "Verification failed", status_code: int = 400) -> None:
        super().__init__(message)
        self.statusCode = status_code


@dataclass
class WebhookPayload:
    event: EventType | str
    expires_at: int
    # Allow arbitrary extra fields
    # mypy: ignore dynamic attributes


# Re-export WebhookMessage for convenience and type checking
# This provides the full typed webhook message structure
__all__ = [
    "VerificationError",
    "WebhookPayload",
    "WebhookVerificationOptions",
    "WebhookMessage",
]


@dataclass
class WebhookVerificationOptions:
    allowExpired: Optional[bool] = None




from __future__ import annotations

from typing import Optional

from .types import *  # re-export types for convenience

def _client():
    # Lazy import to avoid circular imports during package initialization
    from .._default import get_client as _get_client
    return _get_client()

__all__ = [
    "EventType",
    "DryRun",
    "Metadata",
    "RunSpecificWebhook",
    "PayloadWebhook",
    "StartRunRequest",
    "StartRunResponse",
    "UserInteractionData",
    "VideoUrl",
    "FileUrl",
    "ScreenshotUrl",
    "RunError",
    "RunResult",
    "WebhookEvent",
    "WebhookReplayResponse",
    "RunHandle",
    "RunStreamOptions",
    "SseEventName",
    "SseMessage",
    "RunEventEnvelope",
    # Convenience APIs
    "start",
    "subscribe_to_session",
    "submit_user_interaction",
    "get_results",
    "interrupt",
    "replay_webhooks",
]


def start(request: StartRunRequest, options: Optional[RunStreamOptions] = None) -> RunHandle:
    return _client().runs.start(request, options)


def subscribe_to_session(session_id: str, options: Optional[RunStreamOptions] = None) -> RunHandle:
    return _client().runs.subscribe_to_session(session_id, options)


def submit_user_interaction(session_id: str, data: UserInteractionData) -> None:
    return _client().runs.submit_user_interaction(session_id, data)


def get_results(session_id: str) -> RunResult:
    return _client().runs.get_results(session_id)


def interrupt(session_id: str) -> None:
    return _client().runs.interrupt(session_id)


def replay_webhooks(session_id: str) -> WebhookReplayResponse:
    return _client().runs.replay_webhooks(session_id)



from __future__ import annotations

from typing import Any, Dict, Optional

from .types import VaultEntry, GetVaultEntriesFilters, ProxyConfig, VaultPostPutHeadersInBody

def _client():
    # Lazy import to avoid circular imports during package initialization
    from .._default import get_client as _get_client
    return _get_client()

__all__ = [
    "VaultEntry",
    "GetVaultEntriesFilters",
    "ProxyConfig",
    "VaultPostPutHeadersInBody",
    # Convenience APIs
    "create",
    "get",
    "update",
    "delete",
]


def create(
    domain: str,
    permissioned_user_id: str,
    options: Optional[Dict[str, Any]] = None,
) -> VaultEntry:
    return _client().vault.create(domain, permissioned_user_id, options)


def get(filters: Optional[GetVaultEntriesFilters] = None):
    return _client().vault.get(filters)


def update(updates: Dict[str, Any]) -> VaultEntry:
    return _client().vault.update(updates)


def delete(params: Dict[str, str]) -> None:
    return _client().vault.delete(params)



import unittest

from cloudcruise.workflows.client import WorkflowsClient
from cloudcruise.workflows.types import WorkflowMetadata, WorkflowInputSchema, InputValidationError

class _FakeClient(WorkflowsClient):
    def __init__(self):
        super().__init__(lambda *args, **kwargs: None)
        self._meta = WorkflowMetadata(
            input_schema=WorkflowInputSchema(
                type="object",
                properties={
                    "url": {"type": "string"},
                    "count": {"type": ["integer", "null"]},
                },
                required=["url"],
                additionalProperties=False,
            )
        )

    def get_workflow_metadata(self, workflow_id: str):
        return self._meta

class TestWorkflowValidation(unittest.TestCase):
    def test_validate_success(self):
        c = _FakeClient()
        c.validate_workflow_input("wf-1", {"url": "https://example.com", "count": 3})

    def test_validate_missing_required(self):
        c = _FakeClient()
        with self.assertRaises(InputValidationError):
            c.validate_workflow_input("wf-1", {"count": 3})

    def test_validate_unknown_key(self):
        c = _FakeClient()
        with self.assertRaises(InputValidationError):
            c.validate_workflow_input("wf-1", {"url": "x", "extra": 1})

if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

import threading
from typing import Any, Callable, Dict, Optional
import requests
import json
import re


SSEEvent = Dict[str, Any]


class SSEHandlers:
    def __init__(
        self,
        on_open: Optional[Callable[[], None]] = None,
        on_event: Optional[Callable[[SSEEvent], None]] = None,
        on_error: Optional[Callable[[Exception], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
    ) -> None:
        self.on_open = on_open
        self.on_event = on_event
        self.on_error = on_error
        self.on_close = on_close


class SSEConnection:
    def __init__(self, closer: Callable[[], None]) -> None:
        self._close = closer

    def close(self) -> None:
        self._close()


def _parse_frame(frame: str) -> SSEEvent:
    event = "message"
    data = ""
    _id: Optional[str] = None
    for line in frame.splitlines():
        if not line or line.startswith(":"):
            continue
        idx = line.find(":")
        field = line if idx == -1 else line[:idx]
        value = "" if idx == -1 else line[idx + 1 :].lstrip()
        if field == "event":
            event = value
        elif field == "data":
            data = (data + "\n" + value) if data else value
        elif field == "id":
            _id = value
    parsed: Any
    try:
        parsed = None if data == "" else json.loads(data)
    except Exception:
        parsed = data
    return {"event": event, "data": parsed, "id": _id, "raw": frame}


def open_sse(
    url: str,
    handlers: SSEHandlers,
    headers: Optional[Dict[str, str]] = None,
    with_credentials: bool = False,
    stop_event: Optional[threading.Event] = None,
) -> SSEConnection:
    """
    Opens an SSE connection using requests streaming and parses frames.
    """
    session = requests.Session()
    req_headers = {
        "Accept": "text/event-stream",
        "Cache-Control": "no-cache",
    }
    if headers:
        req_headers.update(headers)

    cancelled = threading.Event()
    if stop_event is None:
        stop_event = threading.Event()

    def run() -> None:
        try:
            with session.get(url, headers=req_headers, stream=True, timeout=60) as resp:
                if not resp.ok:
                    raise RuntimeError(f"SSE HTTP {resp.status_code}")
                if handlers.on_open:
                    try:
                        handlers.on_open()
                    except Exception:
                        pass
                buffer = ""
                for chunk in resp.iter_content(chunk_size=None):
                    if stop_event.is_set() or cancelled.is_set():
                        break
                    if not chunk:
                        continue
                    try:
                        text = chunk.decode("utf-8", errors="ignore")
                    except Exception:
                        continue
                    parts = re.split(r"\r?\n\r?\n", buffer + text)
                    buffer = parts.pop() if parts else ""
                    for frame in parts:
                        evt = _parse_frame(frame)
                        if handlers.on_event:
                            try:
                                handlers.on_event(evt)
                            except Exception:
                                pass
        except Exception as e:
            if handlers.on_error:
                try:
                    handlers.on_error(e if isinstance(e, Exception) else Exception(str(e)))
                except Exception:
                    pass
        finally:
            if handlers.on_close:
                try:
                    handlers.on_close()
                except Exception:
                    pass

    thread = threading.Thread(target=run, name="cloudcruise-sse", daemon=True)
    thread.start()

    def _close() -> None:
        cancelled.set()
        stop_event.set()

    return SSEConnection(_close)



from __future__ import annotations

from typing import Any, Dict, List, Optional
from .types import VaultEntry, GetVaultEntriesFilters
from .utils import encrypt_sensitive_fields, decrypt_sensitive_fields


class VaultClient:
    def __init__(self, make_request, encryption_key: str) -> None:
        self._make_request = make_request
        self._encryption_key = encryption_key

    def create(
        self,
        domain: str,
        permissioned_user_id: str,
        options: Optional[Dict[str, Any]] = None,
    ) -> VaultEntry:
        entry: Dict[str, Any] = {
            "domain": domain,
            "permissioned_user_id": permissioned_user_id,
        }
        if options:
            entry.update(options)

        processed = encrypt_sensitive_fields(entry, self._encryption_key)
        response = self._make_request("POST", "/vault", processed)
        return decrypt_sensitive_fields(response, self._encryption_key)

    def get(self, filters: Optional[GetVaultEntriesFilters] = None):
        path = "/vault"
        if filters and (filters.permissioned_user_id or filters.domain):
            from urllib.parse import urlencode

            params: Dict[str, Any] = {}
            if filters.permissioned_user_id:
                params["permissioned_user_id"] = filters.permissioned_user_id
            if filters.domain:
                params["domain"] = filters.domain
            qs = urlencode(params)
            path += f"?{qs}"

        response = self._make_request("GET", path)
        entries = response if isinstance(response, list) else [response]

        should_decrypt = True
        if filters and filters.decryptCredentials is False:
            should_decrypt = False
        if should_decrypt:
            entries = [decrypt_sensitive_fields(e, self._encryption_key) for e in entries]
        return entries

    def update(self, updates: Dict[str, Any]) -> VaultEntry:
        """
        Updates an existing vault entry
        Required fields: permissioned_user_id, user_name, password, domain
        """
        if not updates.get("permissioned_user_id"):
            raise ValueError("permissioned_user_id is required for vault updates")
        if not updates.get("user_name"):
            raise ValueError("user_name is required for vault updates")
        if not updates.get("password"):
            raise ValueError("password is required for vault updates")
        if not updates.get("domain"):
            raise ValueError("domain is required for vault updates")

        processed = encrypt_sensitive_fields(dict(updates), self._encryption_key)
        response = self._make_request("PUT", "/vault", processed)
        return decrypt_sensitive_fields(response, self._encryption_key)

    def delete(self, params: Dict[str, str]) -> None:
        """
        Deletes a vault entry by domain and permissioned_user_id
        params: { "domain": str, "permissioned_user_id": str }
        """
        if not params.get("domain"):
            raise ValueError("domain is required to delete a vault entry")
        if not params.get("permissioned_user_id"):
            raise ValueError("permissioned_user_id is required to delete a vault entry")
        self._make_request("DELETE", "/vault", params)



from __future__ import annotations

import json
from typing import Any, Dict
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
import secrets


def _hex_to_bytes(h: str) -> bytes:
    return bytes.fromhex(h)


def _bytes_to_hex(b: bytes) -> str:
    return b.hex()


def encrypt_data(data: Any, key_hex: str) -> str:
    try:
        key = _hex_to_bytes(key_hex)
        if len(key) != 32:
            raise ValueError("Invalid key length; expected 32-byte (256-bit) key in hex")
        iv = secrets.token_bytes(12)
        aesgcm = AESGCM(key)
        plaintext = json.dumps(data).encode("utf-8")
        ct_tag = aesgcm.encrypt(iv, plaintext, None)  # ciphertext + tag
        # Split tag (last 16 bytes)
        ciphertext, tag = ct_tag[:-16], ct_tag[-16:]
        return _bytes_to_hex(iv) + _bytes_to_hex(ciphertext) + _bytes_to_hex(tag)
    except Exception as e:
        raise RuntimeError(f"Encryption failed: {str(e)}")


def decrypt_data(encrypted_hex: str, key_hex: str) -> Any:
    try:
        if not isinstance(encrypted_hex, str) or len(encrypted_hex) < 56:
            raise ValueError("Invalid encrypted payload")
        key = _hex_to_bytes(key_hex)
        if len(key) != 32:
            raise ValueError("Invalid key length; expected 32-byte (256-bit) key in hex")
        iv = _hex_to_bytes(encrypted_hex[:24])
        tag = _hex_to_bytes(encrypted_hex[-32:])
        ciphertext = _hex_to_bytes(encrypted_hex[24:-32])
        aesgcm = AESGCM(key)
        plaintext = aesgcm.decrypt(iv, ciphertext + tag, None)
        return json.loads(plaintext.decode("utf-8"))
    except Exception as e:
        raise RuntimeError(f"Decryption failed: {str(e)}")


def encrypt_sensitive_fields(entry: Dict[str, Any], encryption_key: str) -> Dict[str, Any]:
    encrypted = dict(entry)
    if entry.get("user_name") is not None:
        encrypted["user_name"] = encrypt_data(entry["user_name"], encryption_key)
    if entry.get("password") is not None:
        encrypted["password"] = encrypt_data(entry["password"], encryption_key)
    if entry.get("tfa_secret") is not None:
        encrypted["tfa_secret"] = encrypt_data(entry["tfa_secret"], encryption_key)
    return encrypted


def decrypt_sensitive_fields(entry: Dict[str, Any], encryption_key: str) -> Dict[str, Any]:
    decrypted = dict(entry)
    val = entry.get("user_name")
    if isinstance(val, str):
        try:
            decrypted["user_name"] = decrypt_data(val, encryption_key)
        except Exception:
            pass
    val = entry.get("password")
    if isinstance(val, str):
        try:
            decrypted["password"] = decrypt_data(val, encryption_key)
        except Exception:
            pass
    val = entry.get("tfa_secret")
    if isinstance(val, str):
        try:
            decrypted["tfa_secret"] = decrypt_data(val, encryption_key)
        except Exception:
            pass
    return decrypted



from __future__ import annotations

from typing import Any, Dict

from .types import (
    Workflow,
    WorkflowInputSchema,
    WorkflowMetadata,
    WorkflowPropertySchema,
    InputValidationError,
)
def _client():
    # Lazy import to avoid circular imports during package initialization
    from .._default import get_client as _get_client
    return _get_client()

__all__ = [
    "Workflow",
    "WorkflowInputSchema",
    "WorkflowMetadata",
    "WorkflowPropertySchema",
    "InputValidationError",
    # Convenience APIs
    "get_all_workflows",
    "get_workflow_metadata",
    "validate_workflow_input",
]


def get_all_workflows() -> list[Workflow]:
    return _client().workflows.get_all_workflows()


def get_workflow_metadata(workflow_id: str) -> WorkflowMetadata:
    return _client().workflows.get_workflow_metadata(workflow_id)


def validate_workflow_input(workflow_id: str, payload: Dict[str, Any]) -> None:
    return _client().workflows.validate_workflow_input(workflow_id, payload)



from __future__ import annotations

import threading
import time
import uuid
from typing import Any, Dict, Optional, Set, Iterator

from .sse import open_sse, SSEHandlers, SSEConnection
from .events import SimpleEventEmitter
from .async_queue import AsyncEventQueue


def _is_final_event(event_type: Optional[str]) -> bool:
    return event_type in {"execution.success", "execution.failed", "execution.stopped"}


class SessionSubscription:
    def __init__(self, emitter: SimpleEventEmitter, queue: AsyncEventQueue[Dict[str, Any]], on_close: callable):
        self._emitter = emitter
        self._queue = queue
        self._on_close = on_close

    def on(self, event: str, handler):
        return self._emitter.on(event, handler)

    def close(self) -> None:
        try:
            self._queue.close()
        finally:
            self._on_close()

    def __iter__(self) -> Iterator[Dict[str, Any]]:
        return iter(self._queue)


class _SessionChannel:
    def __init__(self, session_id: str) -> None:
        self.session_id = session_id
        self.emitter = SimpleEventEmitter()
        self.subscribers: Set[AsyncEventQueue[Dict[str, Any]]] = set()
        self.ended = False


class ConnectionManager:
    def __init__(self, base_url: str, api_key: str) -> None:
        self._base_url = base_url.rstrip("/")
        self._api_key = api_key
        self._client_id: Optional[str] = None
        self._conn: Optional[SSEConnection] = None
        self._connecting = False
        self._connected = False
        self._reconnecting = False
        self._reconnect_delays = [1.0, 3.0, 10.0]
        self._sessions: Dict[str, _SessionChannel] = {}
        self._lock = threading.Lock()

    def ensure_client_id(self) -> str:
        if self._client_id:
            return self._client_id
        self._client_id = str(uuid.uuid4())
        return self._client_id

    def connect_if_needed(self) -> None:
        with self._lock:
            if self._connected or self._connecting:
                return
            if not self._client_id:
                self.ensure_client_id()
            self._open_mux_connection()

    def subscribe(self, session_id: str, stop_event: Optional[threading.Event] = None) -> SessionSubscription:
        # Kick off connection if not already
        try:
            self.connect_if_needed()
        except Exception:
            pass

        with self._lock:
            ch = self._sessions.get(session_id)
            if not ch:
                ch = _SessionChannel(session_id)
                self._sessions[session_id] = ch

            q: AsyncEventQueue[Dict[str, Any]] = AsyncEventQueue()
            ch.subscribers.add(q)

            def _on_close() -> None:
                with self._lock:
                    try:
                        q.close()
                    except Exception:
                        pass
                    ch.subscribers.discard(q)
                    if ch.subscribers.__len__() == 0 and ch.ended:
                        self._sessions.pop(session_id, None)

            return SessionSubscription(ch.emitter, q, _on_close)

    def _emit_all(self, event: str, payload: Any | None = None) -> None:
        for ch in list(self._sessions.values()):
            ch.emitter.emit(event, payload)

    def _open_mux_connection(self) -> None:
        if self._connecting or self._connected:
            return
        if not self._client_id:
            self.ensure_client_id()

        self._connecting = True
        url = f"{self._base_url}/run/clients/{self._client_id}/events"
        headers = {"cc-key": self._api_key}

        def on_open() -> None:
            with self._lock:
                self._connected = True
                self._connecting = False
            self._emit_all("open")

        def on_event(evt: Dict[str, Any]) -> None:
            # Expected events: {event: 'ping'| 'run.event', data: {...}}
            if evt.get("event") == "ping":
                self._emit_all("ping", evt)
                return
            if evt.get("event") == "run.event":
                raw_data = evt.get("data")
                data: Dict[str, Any] = {}
                if isinstance(raw_data, dict):
                    inner = raw_data.get("data")
                    if isinstance(inner, dict):
                        data = inner
                    else:
                        data = raw_data
                payload = data.get("payload") if isinstance(data, dict) else None
                session_id = None
                if isinstance(payload, dict):
                    val = payload.get("session_id") or payload.get("sessionId")
                    if isinstance(val, str):
                        session_id = val
                if session_id is None and isinstance(data, dict):
                    val = data.get("session_id") or data.get("sessionId")
                    if isinstance(val, str):
                        session_id = val
                if not session_id:
                    return
                ch = self._sessions.get(session_id)
                if not ch:
                    return
                msg = {"event": "run.event", "data": data}
                for q in list(ch.subscribers):
                    q.push(msg)
                ch.emitter.emit("run.event", msg)
                ev_type = data.get("event")
                if isinstance(ev_type, str) and _is_final_event(ev_type):
                    ch.ended = True
                    ch.emitter.emit("end", {"type": ev_type})
                    for q in list(ch.subscribers):
                        q.close()
                    ch.subscribers.clear()
                    # Remove channel
                    self._sessions.pop(session_id, None)

        def on_error(err: Exception) -> None:
            self._emit_all("error", err)
            with self._lock:
                if not self._reconnecting:
                    self._schedule_reconnect()

        def on_close() -> None:
            with self._lock:
                self._connected = False
                self._connecting = False
            self._emit_all("close")
            with self._lock:
                if not self._reconnecting:
                    self._schedule_reconnect()

        try:
            self._conn = open_sse(
                url,
                SSEHandlers(on_open=on_open, on_event=on_event, on_error=on_error, on_close=on_close),
                headers=headers,
            )
        except Exception:
            with self._lock:
                self._connected = False
                self._connecting = False
                if not self._reconnecting:
                    self._schedule_reconnect()

    def _schedule_reconnect(self) -> None:
        if self._reconnecting:
            return
        self._reconnecting = True

        def worker() -> None:
            for delay in self._reconnect_delays:
                # Notify listeners about reconnect attempt
                for ch in list(self._sessions.values()):
                    ch.emitter.emit("reconnect", {"attemptDelayMs": int(delay * 1000)})
                time.sleep(delay)
                try:
                    self._open_mux_connection()
                    if self._connected:
                        self._reconnecting = False
                        return
                except Exception:
                    pass
            # Give up; keep flag false so that new subscribe/connect can reattempt
            self._reconnecting = False

        t = threading.Thread(target=worker, name="cloudcruise-reconnect", daemon=True)
        t.start()



from __future__ import annotations

import hmac
import hashlib
import json

from .types import VerificationError, WebhookPayload, WebhookVerificationOptions


def _verify_hmac(received_data: str, received_signature: str, secret_key: str) -> bool:
    mac = hmac.new(secret_key.encode("utf-8"), received_data.encode("utf-8"), hashlib.sha256)
    calculated = mac.hexdigest()
    # received header like: "sha256=<hex>"; split and take hex part if formatted that way
    formatted = received_signature.split("=")[-1]
    if len(formatted) != len(calculated):
        return False
    try:
        return hmac.compare_digest(bytes.fromhex(calculated), bytes.fromhex(formatted))
    except Exception:
        # If not hex-encoded, fallback to string compare_digest
        return hmac.compare_digest(calculated, formatted)


def verify_message(
    raw_body: bytes,
    received_signature: str,
    secret_key: str,
    options: WebhookVerificationOptions | None = None,
) -> WebhookPayload:
    if not raw_body:
        raise VerificationError("Received request without body", 400)
    if not received_signature:
        raise VerificationError("Missing HMAC signature", 400)
    if not secret_key:
        raise VerificationError("Missing secret key", 400)

    try:
        data_string = raw_body.decode("utf-8")
    except UnicodeDecodeError as e:
        raise VerificationError(f"Failed to decode body as UTF-8: {str(e)}", 400)

    try:
        data_json = json.loads(data_string)
    except Exception as e:
        raise VerificationError(f"Failed to decode JSON: {str(e)}", 400)

    expires_at = data_json.get("expires_at")
    if not expires_at:
        raise VerificationError("No expiration date sent", 400)

    if not _verify_hmac(data_string, received_signature, secret_key):
        raise VerificationError("Invalid HMAC signature", 401)

    if not (options and options.allowExpired) and (int(__import__('time').time()) > int(expires_at)):
        raise VerificationError("Webhook message expired", 400)

    return data_json  # type: ignore




from __future__ import annotations

from .types import WebhookPayload, WebhookVerificationOptions, VerificationError, WebhookMessage
from .utils import verify_message as verify_signature

__all__ = [
    "WebhookPayload",
    "WebhookVerificationOptions",
    "VerificationError",
    "WebhookMessage",
    # Convenience function (no client needed)
    "verify_signature",
]



from __future__ import annotations

from .types import (
    ExecutionQueuedPayload,
    ExecutionStartPayload,
    ExecutionStepPayload,
    InteractionWaitingPayload,
    InteractionFinishedPayload,
    InteractionFinishedPayloadVariant1,
    InteractionFinishedPayloadVariant2,
    AgentErrorAnalysisPayload,
    ExecutionRequeuedPayload,
    EndRunPayload,
    EndRunError,
    ExecutionStoppedEarlyPayload,
    FileUploadedPayload,
    ScreenshotUploadedPayload,
    WebhookMessage,
    RunEventMessage,
    RunEventMessageData,
)

__all__ = [
    # Event payload types
    "ExecutionQueuedPayload",
    "ExecutionStartPayload",
    "ExecutionStepPayload",
    "InteractionWaitingPayload",
    "InteractionFinishedPayload",
    "InteractionFinishedPayloadVariant1",
    "InteractionFinishedPayloadVariant2",
    "AgentErrorAnalysisPayload",
    "ExecutionRequeuedPayload",
    "EndRunPayload",
    "EndRunError",
    "ExecutionStoppedEarlyPayload",
    "FileUploadedPayload",
    "ScreenshotUploadedPayload",
    # Generic message types
    "WebhookMessage",
    "RunEventMessage",
    "RunEventMessageData",
]



from __future__ import annotations

import threading
from typing import Generic, Iterator, Optional, TypeVar


T = TypeVar("T")


class AsyncEventQueue(Generic[T]):
    """
    A thread-cooperative queue that supports iteration until closed.
    """

    def __init__(self) -> None:
        self._items: list[T] = []
        self._closed = False
        self._cv = threading.Condition()

    def push(self, item: T) -> None:
        with self._cv:
            if self._closed:
                return
            self._items.append(item)
            self._cv.notify()

    def close(self) -> None:
        with self._cv:
            if self._closed:
                return
            self._closed = True
            self._cv.notify_all()

    def __iter__(self) -> Iterator[T]:
        while True:
            with self._cv:
                while not self._items and not self._closed:
                    self._cv.wait()
                if self._items:
                    item = self._items.pop(0)
                    yield item
                elif self._closed:
                    return




from __future__ import annotations

import os
from typing import Optional
from typing import Literal


CloudCruiseEnvVar = Literal[
    "CLOUDCRUISE_API_KEY",
    "CLOUDCRUISE_BASE_URL",
    "CLOUDCRUISE_ENCRYPTION_KEY",
]


def get_env(key: CloudCruiseEnvVar) -> Optional[str]:
    return os.environ.get(key)



from __future__ import annotations

from typing import Callable, Dict, Set, Any


EventHandler = Callable[[Any], None]


class SimpleEventEmitter:
    def __init__(self) -> None:
        self._listeners: Dict[str, Set[EventHandler]] = {}

    def on(self, event: str, handler: EventHandler) -> Callable[[], None]:
        if event not in self._listeners:
            self._listeners[event] = set()
        self._listeners[event].add(handler)

        def off() -> None:
            s = self._listeners.get(event)
            if s and handler in s:
                s.remove(handler)

        return off

    def emit(self, event: str, payload: Any | None = None) -> None:
        handlers = self._listeners.get(event)
        if not handlers:
            return
        for h in list(handlers):
            try:
                h(payload)
            except Exception:
                # Swallow exceptions in user handlers to avoid breaking emitter
                pass

    def clear(self) -> None:
        self._listeners.clear()




from .cloudcruise import CloudCruise, CloudCruiseParams

from .vault.types import (
    VaultEntry,
    GetVaultEntriesFilters,
    ProxyConfig,
    VaultPostPutHeadersInBody,
)

from .workflows.types import (
    Workflow,
    WorkflowInputSchema,
    WorkflowMetadata,
    InputValidationError,
)

from .runs.types import (
    EventType,
    DryRun,
    Metadata,
    RunSpecificWebhook,
    PayloadWebhook,
    StartRunRequest,
    StartRunResponse,
    UserInteractionData,
    VideoUrl,
    FileUrl,
    ScreenshotUrl,
    RunError,
    RunResult,
    WebhookEvent,
    WebhookReplayResponse,
    RunHandle,
    RunStreamOptions,
    SseEventName,
    SseMessage,
    RunEventEnvelope,
    # Event payload types
    ExecutionQueuedPayload,
    ExecutionStartPayload,
    ExecutionStepPayload,
    InteractionWaitingPayload,
    InteractionFinishedPayload,
    AgentErrorAnalysisPayload,
    ExecutionRequeuedPayload,
    EndRunPayload,
    EndRunError,
    ExecutionStoppedEarlyPayload,
    FileUploadedPayload,
    ScreenshotUploadedPayload,
    EventWebhookMessage,
    RunEventMessage,
)

from .webhook.types import WebhookPayload, WebhookVerificationOptions, VerificationError, WebhookMessage
from ._default import get_client as client

__all__ = [
    "CloudCruise",
    "CloudCruiseParams",
    # Default client helper
    "client",
    # Vault Types
    "VaultEntry",
    "GetVaultEntriesFilters",
    "ProxyConfig",
    "VaultPostPutHeadersInBody",
    # Workflow Types
    "Workflow",
    "WorkflowInputSchema",
    "WorkflowMetadata",
    "InputValidationError",
    # Run Types
    "EventType",
    "DryRun",
    "Metadata",
    "RunSpecificWebhook",
    "PayloadWebhook",
    "StartRunRequest",
    "StartRunResponse",
    "UserInteractionData",
    "VideoUrl",
    "FileUrl",
    "ScreenshotUrl",
    "RunError",
    "RunResult",
    "WebhookEvent",
    "WebhookReplayResponse",
    "RunHandle",
    "RunStreamOptions",
    "SseEventName",
    "SseMessage",
    "RunEventEnvelope",
    # Event Payload Types
    "ExecutionQueuedPayload",
    "ExecutionStartPayload",
    "ExecutionStepPayload",
    "InteractionWaitingPayload",
    "InteractionFinishedPayload",
    "AgentErrorAnalysisPayload",
    "ExecutionRequeuedPayload",
    "EndRunPayload",
    "EndRunError",
    "ExecutionStoppedEarlyPayload",
    "FileUploadedPayload",
    "ScreenshotUploadedPayload",
    "EventWebhookMessage",
    "RunEventMessage",
    # Webhook Types
    "WebhookPayload",
    "WebhookVerificationOptions",
    "VerificationError",
    "WebhookMessage",
]



import json
import time
import hmac
import hashlib
import unittest

from cloudcruise.webhook.client import WebhookClient

def _sign(body: str, secret: str) -> str:
    mac = hmac.new(secret.encode(), body.encode(), hashlib.sha256).hexdigest()
    return f"sha256={mac}"

class TestWebhook(unittest.TestCase):
    def test_verify_signature_ok(self):
        client = WebhookClient()
        body = {"event": "execution.success", "expires_at": int(time.time()) + 60, "x": 1}
        body_str = json.dumps(body)
        body_bytes = body_str.encode("utf-8")
        sig = _sign(body_str, "sekrit")
        verified = client.verify_signature(body_bytes, sig, "sekrit")
        self.assertEqual(verified["event"], "execution.success")
        self.assertEqual(verified["x"], 1)

    def test_verify_signature_expired_rejected(self):
        client = WebhookClient()
        body = {"event": "execution.success", "expires_at": int(time.time()) - 1, "x": 1}
        body_str = json.dumps(body)
        body_bytes = body_str.encode("utf-8")
        sig = _sign(body_str, "sekrit")
        with self.assertRaises(Exception) as ctx:
            client.verify_signature(body_bytes, sig, "sekrit")
        self.assertIn("expired", str(ctx.exception).lower())

if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Union


@dataclass
class Workflow:
    id: str
    name: str
    description: Optional[str]
    created_at: str
    updated_at: str
    workspace_id: str
    created_by: str
    enable_popup_handling: bool
    enable_xpath_recovery: bool
    enable_error_code_generation: bool
    enable_service_unavailable_recovery: bool
    enable_action_timing_recovery: bool


WorkflowPropertySchema = Union[str, List[str], Dict[str, Any]]


@dataclass
class WorkflowInputSchema:
    type: Optional[str] = None
    properties: Optional[Dict[str, WorkflowPropertySchema]] = None
    required: Optional[List[str]] = None
    additionalProperties: Optional[bool] = None


@dataclass
class WorkflowMetadata:
    input_schema: WorkflowInputSchema


@dataclass
class InvalidTypeDetail:
    field: str
    expected_display: str
    actual: str


class InputValidationError(Exception):
    def __init__(
        self,
        message: str = "Input validation failed",
        missing_required: Optional[List[str]] = None,
        invalid_types: Optional[List[InvalidTypeDetail]] = None,
        unknown_keys: Optional[List[str]] = None,
    ) -> None:
        super().__init__(message)
        self.missingRequired = missing_required or []
        self.invalidTypes = invalid_types or []
        self.unknownKeys = unknown_keys or []




from __future__ import annotations

import threading
import time
from typing import Any, Dict, Iterator, Optional

from ..utils.async_queue import AsyncEventQueue
from ..utils.events import SimpleEventEmitter
from ..utils.connection_manager import ConnectionManager, SessionSubscription
from ..workflows.client import WorkflowsClient
from .types import (
    StartRunRequest,
    UserInteractionData,
    RunResult,
    WebhookReplayResponse,
    RunStreamOptions,
    SseMessage,
    RunHandle,
)

class RunsClient:
    def __init__(
        self,
        connection_manager: ConnectionManager,
        make_request,
        workflows: Optional[WorkflowsClient] = None,
    ) -> None:
        self._make_request = make_request
        self._workflows = workflows
        self._connection_manager = connection_manager

    def start(self, request: StartRunRequest, options: Optional[RunStreamOptions] = None) -> RunHandle:
        if self._workflows is not None:
            # Validate input variables proactively
            self._workflows.validate_workflow_input(request.workflow_id, request.run_input_variables)

        client_id = self._connection_manager.ensure_client_id()
        self._connection_manager.connect_if_needed()
        request.client_id = client_id
        from dataclasses import is_dataclass, asdict
        payload = asdict(request) if is_dataclass(request) else (
            dict(request) if isinstance(request, dict) else request.__dict__
        )
        resp = self._make_request("POST", "/run", payload)
        session_id: Optional[str]
        if isinstance(resp, dict):
            session_id = resp.get("session_id") or resp.get("sessionId")
        else:
            session_id = getattr(resp, "session_id", None) or getattr(resp, "sessionId", None)
        if not session_id:
            raise RuntimeError("CloudCruise start run response did not include session_id")
        return self.subscribe_to_session(session_id, options)

    def subscribe_to_session(self, session_id: str, options: Optional[RunStreamOptions] = None) -> RunHandle:
        emitter = SimpleEventEmitter()
        stream: AsyncEventQueue[SseMessage] = AsyncEventQueue()

        ended = False
        closed = False
        sub: Optional[SessionSubscription] = None

        reconnect_enabled = True if options is None or options.reconnect_enabled is None else options.reconnect_enabled
        reconnect_delays = options.reconnect_delays if options and options.reconnect_delays else [1.0, 3.0, 10.0]

        def is_terminal(status: Optional[str]) -> bool:
            return status in {"execution.success", "execution.failed", "execution.stopped"}

        def flatten_event(msg: Dict[str, Any]) -> Dict[str, Any]:
            """
            Flatten the nested SSE event structure for better UX.
            Transforms:
              {
                'event': 'run.event',
                'data': {
                  'event': 'execution.start',
                  'payload': {...},
                  'timestamp': ...
                }
              }
            Into:
              {
                'type': 'execution.start',
                'payload': {...},
                'timestamp': ...,
                '_raw': {...}  # original message
              }
            """
            data = msg.get("data", {})
            flattened = {
                "type": data.get("event"),
                "payload": data.get("payload", {}),
                "timestamp": data.get("timestamp"),
                "expires_at": data.get("expires_at"),
                "_raw": msg,
            }
            return flattened

        def emit(event: str, payload: Any | None = None) -> None:
            emitter.emit(event, payload)
            if event in ("run.event", "ping"):
                emitter.emit("message", payload)

        def end_and_cleanup(status: str) -> None:
            nonlocal ended, closed, sub
            if ended:
                return
            ended = True
            closed = True
            try:
                if sub is not None:
                    sub.close()
            except Exception:
                pass
            emit("end", {"type": status})
            stream.close()
            emitter.clear()

        def connect() -> None:
            nonlocal sub
            sub = self._connection_manager.subscribe(session_id)
            s = sub
            s.on("open", lambda _=None: emit("open"))
            s.on("ping", lambda evt: emit("ping", evt))

            def on_run_event(msg: Any) -> None:
                nonlocal ended
                m = msg  # expected dict
                if not isinstance(m, dict) or m.get("event") != "run.event":
                    return

                # Flatten the event for better UX
                flattened = flatten_event(m)
                event_type = flattened.get("type")

                # Push original message to stream for iteration
                stream.push(m)  # type: ignore

                # Emit flattened event to 'run.event' listeners
                emit("run.event", flattened)

                # Also emit to type-specific listeners (e.g., 'execution.start')
                # This matches the JS SDK behavior and provides better DX
                if event_type and isinstance(event_type, str):
                    try:
                        emit(event_type, flattened)
                    except Exception:
                        pass  # Ignore errors in type-specific emission

                if isinstance(event_type, str) and is_terminal(event_type):
                    end_and_cleanup(event_type)

            s.on("run.event", on_run_event)
            s.on("error", lambda err: _on_error(err))
            s.on("reconnect", lambda e: emit("reconnect", e))
            s.on("end", lambda e: end_and_cleanup((e or {}).get("type", "execution.stopped")))

        def _on_error(err: Any) -> None:
            emit("error", err)
            if not reconnect_enabled or ended or closed:
                return
            def worker():
                for base in reconnect_delays:
                    if ended or closed:
                        return
                    time.sleep(base)
                    if ended or closed:
                        return
                    try:
                        snapshot = self.get_results(session_id)
                        status = snapshot.get("status") if isinstance(snapshot, dict) else snapshot.status
                        if isinstance(status, str) and is_terminal(status):
                            end_and_cleanup(status)
                            return
                    except Exception:
                        pass
                    emit("reconnect", {"attemptDelayMs": int(base * 1000)})
                    return  # Connection manager handles reconnect of mux
            t = threading.Thread(target=worker, name="cloudcruise-run-reconnect", daemon=True)
            t.start()

        connect()

        client = self

        class _RunHandle:
            sessionId = session_id

            def on(self, event: str, handler):
                return emitter.on(event, handler)

            def wait(self) -> RunResult:
                # Block until end and then fetch results
                if ended:
                    return client.get_results(session_id)

                done = threading.Event()
                result_container: Dict[str, Any] = {}

                def on_end(_):
                    try:
                        result_container["result"] = client.get_results(session_id)
                    finally:
                        done.set()

                def on_error(err):
                    result_container["error"] = err
                    done.set()

                off_end = self.on("end", on_end)
                off_err = self.on("error", on_error)
                done.wait()
                try:
                    if "error" in result_container:
                        err = result_container["error"]
                        raise err if isinstance(err, Exception) else RuntimeError(f"SSE error: {err}")
                    return result_container["result"]
                finally:
                    try:
                        off_end()
                    except Exception:
                        pass
                    try:
                        off_err()
                    except Exception:
                        pass

            def close(self) -> None:
                nonlocal closed, sub
                closed = True
                try:
                    if sub is not None:
                        sub.close()
                except Exception:
                    pass
                stream.close()
                emitter.clear()

            def __iter__(self) -> Iterator[SseMessage]:
                for msg in stream:
                    yield msg

        return _RunHandle()

    def submit_user_interaction(self, session_id: str, data: UserInteractionData) -> None:
        path = f"/run/{session_id}/user_interaction"
        self._make_request("POST", path, data)

    def get_results(self, session_id: str) -> RunResult:
        path = f"/run/{session_id}"
        return self._make_request("GET", path)

    def interrupt(self, session_id: str) -> None:
        path = f"/run/{session_id}/interrupt"
        self._make_request("POST", path)

    def replay_webhooks(self, session_id: str) -> WebhookReplayResponse:
        path = f"/webhooks/{session_id}/replay"
        return self._make_request("POST", path)



from __future__ import annotations

from typing import Any, Dict, Generic, List, Literal, TypeVar, TypedDict, Union

# Re-use EventType from runs.types to avoid circular imports
# We'll import it later after updating runs/types.py

# Event Payload TypedDicts
class ExecutionQueuedPayload(TypedDict):
    session_id: str
    workflow_id: str


class ExecutionStartPayload(TypedDict, total=False):
    session_id: str
    workflow_id: str
    live_view_url: str  # optional


class ExecutionStepPayload(TypedDict):
    session_id: str
    workflow_id: str
    current_step: str
    next_step: str


class InteractionWaitingPayload(TypedDict):
    session_id: str
    workflow_id: str
    current_step: str
    missing_properties: List[str]
    expected_json_schema_datamodel: Dict[str, Any]
    message: str


# InteractionFinishedPayload has two possible shapes
class InteractionFinishedPayloadVariant1(TypedDict):
    session_id: str
    workflow_id: str
    current_step: str
    missing_properties: List[str]  # empty list in this variant
    expected_json_schema_datamodel: Dict[str, Any]
    message: str


class InteractionFinishedPayloadVariant2(TypedDict, total=False):
    session_id: str
    workflow_id: str
    provided_input: Any
    message: str  # optional
    expected_json_schema_datamodel: Dict[str, Any]


InteractionFinishedPayload = Union[
    InteractionFinishedPayloadVariant1,
    InteractionFinishedPayloadVariant2
]


class AgentErrorAnalysisPayload(TypedDict, total=False):
    analysis_step_name: str
    ai_analysis: str  # optional
    root_cause_analysis: str  # optional
    error_category: str  # optional


class ExecutionRequeuedPayload(TypedDict, total=False):
    session_id: str
    workflow_id: str
    retry_attempt: int
    max_retries: int  # optional
    next_execution_time: str
    delay_ms: int


class EndRunError(TypedDict, total=False):
    message: str
    error_id: str
    full_url: str  # optional
    created_at: str
    error_code: str  # optional
    action_type: str  # optional
    action_display_name: str  # optional
    llm_error_category: str  # optional


class EndRunPayload(TypedDict, total=False):
    session_id: str
    workflow_id: str
    data: Any
    input_variables: Dict[str, Any]
    errors: List[EndRunError]
    status: Literal["execution.success", "execution.failed", "execution.stopped"]
    encrypted_variables: Union[List[str], None]
    file_urls: Union[List[Any], None]


class ExecutionStoppedEarlyPayload(TypedDict):
    message: str
    error_code: str
    session_id: str


class FileUploadedPayload(TypedDict):
    signed_file_url: str
    file_name: str
    timestamp: str
    signed_file_url_expires: str
    metadata: Dict[str, Any]
    session_id: str


class ScreenshotUploadedPayload(TypedDict):
    screenshot_id: str
    signed_screenshot_url: str
    node_display_name: str
    node_id: str
    timestamp: str
    signed_screenshot_url_expires: str
    error_screenshot: bool
    retry_index: int
    full_length_screenshot: bool
    session_id: str


# Type variable for generic webhook messages
E = TypeVar('E', bound=str)


# Generic Webhook Message type
class WebhookMessage(TypedDict, Generic[E], total=False):
    event: E
    timestamp: int
    expires_at: int
    payload: Any  # Will be typed more specifically with EventPayloadMap
    metadata: Dict[str, Any]  # optional


# Generic SSE Run Event Message type
class RunEventMessageData(TypedDict, Generic[E]):
    event: E
    payload: Any
    timestamp: int
    expires_at: int


class RunEventMessage(TypedDict, Generic[E]):
    event: Literal["run.event"]
    data: RunEventMessageData[E]
    timestamp: str
    expires_at: str



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Iterator, List, Literal, Optional, Protocol, TypedDict, Union

# Import event payload types for re-export
from ..events.types import (
    ExecutionQueuedPayload,
    ExecutionStartPayload,
    ExecutionStepPayload,
    InteractionWaitingPayload,
    InteractionFinishedPayload,
    AgentErrorAnalysisPayload,
    ExecutionRequeuedPayload,
    EndRunPayload,
    EndRunError,
    ExecutionStoppedEarlyPayload,
    FileUploadedPayload,
    ScreenshotUploadedPayload,
    WebhookMessage as EventWebhookMessage,
    RunEventMessage,
)


EventType = Literal[
    "execution.queued",
    "execution.start",
    "execution.step",
    "execution.pause",
    "execution.stopped",
    "execution.failed",
    "execution.success",
    "execution.requeued",
    "file.uploaded",
    "screenshot.uploaded",
    "video.uploaded",
    "interaction.waiting",
    "interaction.finished",
    "interaction.failed",
    "agent.error_analysis",
]


@dataclass
class DryRun:
    enabled: bool
    add_to_output: Optional[Dict[str, Any]] = None


@dataclass
class Metadata:
    metadata: Dict[str, Any]


@dataclass
class RunSpecificWebhook:
    url: str
    event_types_subscribed: List[EventType]
    secret: str
    validity: int


PayloadWebhook = Union[Metadata, RunSpecificWebhook]


@dataclass
class StartRunRequest:
    workflow_id: str
    run_input_variables: Dict[str, Any]
    dry_run: Optional[DryRun] = None
    webhook: Optional[PayloadWebhook] = None
    additional_context: Optional[Dict[str, Any]] = None
    client_id: Optional[str] = None


@dataclass
class StartRunResponse:
    session_id: str


UserInteractionData = Dict[str, Any]


@dataclass
class VideoUrl:
    timestamp: str
    session_id: str
    signed_screen_recording_url: str
    signed_screen_recording_url_expires: str


@dataclass
class FileUrl:
    signed_file_url: str
    file_name: str
    timestamp: str
    signed_file_url_expires: str
    metadata: Dict[str, Any]


@dataclass
class ScreenshotUrl:
    signed_screenshot_url: str
    node_display_name: str
    timestamp: str
    signed_screenshot_url_expires: str
    error_screenshot: bool


@dataclass
class RunError:
    prompt: Optional[str] = None
    message: Optional[str] = None
    error_id: Optional[str] = None
    full_url: Optional[str] = None
    llm_model: Optional[str] = None
    created_at: Optional[str] = None
    error_code: Optional[str] = None
    action_type: Optional[str] = None
    action_display_name: Optional[str] = None


@dataclass
class RunResult:
    session_id: str
    status: EventType
    input_variables: Dict[str, Any]
    data: Dict[str, Any]
    video_urls: List[VideoUrl]
    file_urls: List[FileUrl]
    screenshot_urls: List[ScreenshotUrl]
    errors: Optional[List[RunError]]


@dataclass
class WebhookEvent:
    success: bool
    response: str
    error: str


@dataclass
class WebhookReplayResponse:
    status: str
    info: str
    nr_success: int
    nr_failed: int
    webhook_events: List[WebhookEvent]


SseEventName = Literal["run.event", "ping"]


class RunEventData(TypedDict, total=False):
    event: Union[EventType, str]
    payload: Dict[str, Any]
    expires_at: int
    timestamp: int


class RunEventEnvelope(TypedDict, total=False):
    event: Literal["run.event"]
    data: RunEventData
    timestamp: str
    expires_at: str


class PingEnvelope(TypedDict, total=False):
    event: Literal["ping"]
    data: Union[Dict[str, Any], Dict[str, int]]


SseMessage = Union[RunEventEnvelope, PingEnvelope]


@dataclass
class RunStreamOptions:
    # In Python, use a threading.Event to signal stop if desired
    # headers and with_credentials are not used directly in SSE manager
    reconnect_enabled: Optional[bool] = None
    reconnect_delays: Optional[List[float]] = None


class RunHandle(Protocol):
    sessionId: str

    def on(self, event: str, handler) -> Any:
        ...

    def wait(self) -> RunResult:
        ...

    def close(self) -> None:
        ...

    def __iter__(self) -> Iterator[SseMessage]:
        ...


# Export all types including event payloads
__all__ = [
    # Core types
    "EventType",
    "DryRun",
    "Metadata",
    "RunSpecificWebhook",
    "PayloadWebhook",
    "StartRunRequest",
    "StartRunResponse",
    "UserInteractionData",
    "VideoUrl",
    "FileUrl",
    "ScreenshotUrl",
    "RunError",
    "RunResult",
    "WebhookEvent",
    "WebhookReplayResponse",
    "SseEventName",
    "RunEventData",
    "RunEventEnvelope",
    "PingEnvelope",
    "SseMessage",
    "RunStreamOptions",
    "RunHandle",
    # Event payload types (re-exported from events.types)
    "ExecutionQueuedPayload",
    "ExecutionStartPayload",
    "ExecutionStepPayload",
    "InteractionWaitingPayload",
    "InteractionFinishedPayload",
    "AgentErrorAnalysisPayload",
    "ExecutionRequeuedPayload",
    "EndRunPayload",
    "EndRunError",
    "ExecutionStoppedEarlyPayload",
    "FileUploadedPayload",
    "ScreenshotUploadedPayload",
    "EventWebhookMessage",
    "RunEventMessage",
]




from __future__ import annotations

from typing import Optional

from .utils import verify_message
from .types import WebhookPayload, WebhookVerificationOptions


class WebhookClient:
    def __init__(self) -> None:
        pass

    def verify_signature(
        self,
        raw_body: bytes,
        received_signature: str,
        secret_key: str,
        options: Optional[WebhookVerificationOptions] = None,
    ) -> WebhookPayload:
        return verify_message(raw_body, received_signature, secret_key, options)




import unittest

from cloudcruise.vault.utils import encrypt_data, decrypt_data

class TestVaultCrypto(unittest.TestCase):
    def test_encrypt_decrypt_roundtrip(self):
        key_hex = "a" * 64  # 32 bytes hex
        payload = {"user": "alice", "pwd": "secret", "n": 123}

        enc = encrypt_data(payload, key_hex)
        self.assertIsInstance(enc, str)
        self.assertGreater(len(enc), 56)

        dec = decrypt_data(enc, key_hex)
        self.assertEqual(dec, payload)

if __name__ == "__main__":
    unittest.main()



from __future__ import annotations

from typing import Optional

from .cloudcruise import CloudCruise, CloudCruiseParams

_default_client: Optional[CloudCruise] = None


def get_client(
    params: Optional[CloudCruiseParams] = None,
    *,
    api_key: Optional[str] = None,
    encryption_key: Optional[str] = None,
    base_url: Optional[str] = None,
) -> CloudCruise:
    """Return the shared default CloudCruise client.

    You can configure it by passing a CloudCruiseParams instance or via
    keyword args (api_key, encryption_key, base_url). Passing any params or
    keyword args replaces the cached client.
    """
    global _default_client

    effective_params = params
    if effective_params is None and (api_key is not None or encryption_key is not None or base_url is not None):
        effective_params = CloudCruiseParams(api_key=api_key, encryption_key=encryption_key, base_url=base_url)

    if effective_params is not None or _default_client is None:
        _default_client = CloudCruise(effective_params)
    return _default_client



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional


@dataclass
class VaultPostPutHeadersInBody:
    header_name: str
    header_value: str


@dataclass
class ProxyConfig:
    enable: Optional[bool] = None
    target_ip: Optional[str] = None


@dataclass
class VaultEntry:
    id: Optional[str]
    domain: str
    permissioned_user_id: str
    workspace_id: Optional[str] = None
    user_id: Optional[str] = None
    password: Optional[str] = None
    user_name: Optional[str] = None
    tfa_secret: Optional[str] = None
    user_agent: Optional[str] = None
    user_alias: Optional[str] = None
    location: Optional[str] = None
    ip_address: Optional[str] = None
    session_id: Optional[str] = None
    allow_multiple_sessions: Optional[bool] = None
    cookies: Optional[Any] = None
    local_storage: Optional[Any] = None
    session_storage: Optional[Any] = None
    persist_cookies: Optional[bool] = None
    persist_local_storage: Optional[bool] = None
    persist_session_storage: Optional[bool] = None
    cookie_domain_to_store: Optional[str] = None
    proxy: Optional[ProxyConfig] = None
    proxy_string: Optional[str] = None
    headers: Optional[List[VaultPostPutHeadersInBody]] = None
    created_at: Optional[str] = None


@dataclass
class GetVaultEntriesFilters:
    permissioned_user_id: Optional[str] = None
    domain: Optional[str] = None
    decryptCredentials: Optional[bool] = None




from __future__ import annotations

from typing import Any, Dict, List

from .types import (
    Workflow,
    WorkflowMetadata,
    WorkflowInputSchema,
    WorkflowPropertySchema,
    InputValidationError,
    InvalidTypeDetail,
)

class WorkflowsClient:
    def __init__(self, make_request) -> None:
        self._make_request = make_request

    def get_all_workflows(self) -> List[Workflow]:
        return self._make_request("GET", "/workflows")

    def get_workflow_metadata(self, workflow_id: str) -> WorkflowMetadata:
        path = f"/workflows/{workflow_id}/metadata"
        return self._make_request("GET", path)

    def validate_workflow_input(self, workflow_id: str, payload: Dict[str, Any]) -> None:
        meta = self.get_workflow_metadata(workflow_id)

        raw_schema: Any = None
        if isinstance(meta, dict):
            if "input_schema" in meta:
                raw_schema = meta.get("input_schema")
            else:
                m = meta.get("metadata") if isinstance(meta.get("metadata"), dict) else None
                if isinstance(m, dict):
                    raw_schema = m.get("input_schema")
        else:
            try:
                raw_schema = getattr(meta, "input_schema", None)
            except Exception:
                raw_schema = None

        if isinstance(raw_schema, dict):
            properties = raw_schema.get("properties") or {}
            required = raw_schema.get("required") or []
            disallow_extras = (raw_schema.get("additionalProperties") is False)
        else:
            schema: WorkflowInputSchema = raw_schema or WorkflowInputSchema()
            properties = schema.properties or {}
            required = schema.required or []
            disallow_extras = (schema.additionalProperties is False)

        missing_required = [k for k in required if k not in payload]

        def detect_type(v: Any) -> str:
            if v is None:
                return "null"
            if isinstance(v, list):
                return "array"
            if isinstance(v, bool):
                return "boolean"
            if isinstance(v, int) and not isinstance(v, bool):
                return "integer"
            if isinstance(v, float):
                return "number"
            if isinstance(v, dict):
                return "object"
            return "string" if isinstance(v, str) else type(v).__name__

        def expected_types_of(defn: WorkflowPropertySchema) -> List[str]:
            if defn is None:
                return []
            raw = defn
            if isinstance(defn, dict):
                raw = defn.get("type")  # type: ignore
            if raw is None:
                return []
            arr = raw if isinstance(raw, list) else [raw]
            allowed = {"array", "boolean", "integer", "number", "object", "string", "null"}
            out: List[str] = []
            for t in arr:  # type: ignore
                if isinstance(t, str):
                    t = t.lower()
                    if t in allowed:
                        out.append(t)
            return out

        def matches(expected: List[str], actual: str) -> bool:
            if not expected:
                return True
            if actual in expected:
                return True
            if actual == "integer" and "number" in expected:
                return True
            return False

        invalid_types: List[InvalidTypeDetail] = []
        for key, schema_def in properties.items():
            if key not in payload:
                continue
            exp = expected_types_of(schema_def)
            act = detect_type(payload.get(key))
            if not matches(exp, act):
                invalid_types.append(InvalidTypeDetail(field=key, expected_display=" | ".join(exp or ["any"]), actual=act))

        unknown_keys: List[str] = []
        if disallow_extras:
            unknown_keys = [k for k in payload.keys() if k not in properties]

        if missing_required or invalid_types or unknown_keys:
            parts: List[str] = []
            if missing_required:
                parts.append(f"missing required: {', '.join(missing_required)}")
            if invalid_types:
                parts.append("; ".join([f"{e.field}: expected {e.expected_display}, got {e.actual}" for e in invalid_types]))
            if unknown_keys:
                parts.append(f"unknown keys: {', '.join(unknown_keys)}")
            msg = f"Workflow input validation failed: {' | '.join(parts)}"
            raise InputValidationError(msg, missing_required, invalid_types, unknown_keys)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional
import json
import requests

from .utils.env import get_env
from .vault.client import VaultClient
from .workflows.client import WorkflowsClient
from .runs.client import RunsClient
from .webhook.client import WebhookClient
from .utils.connection_manager import ConnectionManager

@dataclass
class CloudCruiseParams:
    api_key: Optional[str] = None
    base_url: Optional[str] = None
    encryption_key: Optional[str] = None


class CloudCruise:
    """
    CloudCruise Python SDK
    Official client library for the CloudCruise Platform
    """

    # Expose typed attributes for IDE autocomplete
    vault: VaultClient
    workflows: WorkflowsClient
    runs: RunsClient
    webhook: WebhookClient

    def __init__(self, params: Optional[CloudCruiseParams] = None) -> None:
        params = params or CloudCruiseParams()
        api_key = params.api_key or get_env("CLOUDCRUISE_API_KEY")
        base_url = params.base_url or get_env("CLOUDCRUISE_BASE_URL") or "https://api.cloudcruise.com"
        encryption_key = params.encryption_key or get_env("CLOUDCRUISE_ENCRYPTION_KEY")

        if not api_key:
            raise ValueError("Missing apiKey. Provide via params.api_key or CLOUDCRUISE_API_KEY env var.")
        if not encryption_key:
            raise ValueError(
                "Missing encryptionKey. Provide via params.encryption_key or CLOUDCRUISE_ENCRYPTION_KEY env var."
            )

        self._api_key = api_key
        self._base_url = base_url.rstrip("/")
        self._encryption_key = encryption_key

        # Initialize namespace clients
        self._connection_manager = ConnectionManager(self._base_url, self._api_key)
        self.vault = VaultClient(self._make_request, self._encryption_key)
        self.workflows = WorkflowsClient(self._make_request)
        self.runs = RunsClient(self._connection_manager, self._make_request, self.workflows)
        self.webhook = WebhookClient()

    def _make_request(self, method: str, path: str, body: Optional[Any] = None) -> Any:
        """
        Makes an HTTP request to the CloudCruise API
        Automatically adds the cc-key header for authentication
        """
        url = f"{self._base_url}{path}"
        headers = {
            "cc-key": self._api_key,
        }
        try:
            # Only send Content-Type when we have a JSON body
            if body is not None:
                headers["Content-Type"] = "application/json"

            resp = requests.request(
                method=method,
                url=url,
                headers=headers,
                data=(json.dumps(body) if body is not None else None),
                timeout=60,
            )
            if not resp.ok:
                error_text = resp.text
                error_message = f"HTTP {resp.status_code}: {resp.reason}"
                try:
                    error_json = resp.json()
                    error_message = error_json.get("message") or error_json.get("error") or error_message
                except Exception:
                    pass
                raise RuntimeError(error_message)

            ctype = resp.headers.get("content-type", "")
            if "application/json" in ctype:
                return resp.json()

            else:
                return resp.text
        except Exception as e:
            raise e
